/* MuSE - Multiple Streaming Engine
 * Copyright (C) 2000-2002 Denis Roio aka jaromil <jaromil@dyne.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id:
 *
 */

#include <in_mpeg.h>
#include <jutils.h>
#include <generic.h>
#include <config.h>


MpegChannel::MpegChannel() : Channel() {
  func("MpegChannel::MpegChannel()");
  type = MP3CHAN;

  loader = NULL;
  server = NULL;
}

MpegChannel::~MpegChannel() {
  func("MpegChannel::~MpegChannel()");
  if(on) stop();
}

/* cleans decoder memory but does'nt kills the thread */
void MpegChannel::clean() {
  if(loader) {
    delete loader;
    loader = NULL;
  }

  if(server) {
    delete server;
    server = NULL;
  }
  opened = false;
}

int MpegChannel::set(char *file) {
  func("MpegChannel::set(%s)",file);
  int res = 0;
  lock();
  on = false;
  
  if(opened)
    clean();
  
  loader = Soundinputstream::hopen(file,&res);
  if(loader == NULL) {
    warning("MpegChannel::set : i don't see any stream flowing from %s",file);
    //    clean();
    unlock();
    return(0);
  } // else func("MpegChannel->loader %p - OK",loader);

  server = new Mpegtoraw(loader);
  if(server == NULL) {
    warning("MpegChannel::set(%s) - uuops! NULL Mpegtoraw",file);
    clean();
    unlock();
    return(0);
  } // else func("MpegChannel->server %p - OK",server);

  server->initialize(file);

  samplerate = server->getfrequency();
  channels = (server->getmode()<3) ? 2 : 1;

  bitrate = server->getbitrate();

  if(!set_mixer()) {
    error("MpegChannel::set(%s) - argh! set_mixer() reported failure",file);
    clean(); unlock(); return(0);
  }

  /* pcm position */
  framepos = 0;
  /* time position */
  time.h = time.m = time.s = 0;
  /* floating point position */
  position = time.f = 0.0;

  /* check if seekable */
  res = (loader->seekable) ? 1: 2;
  seekable = (loader->seekable) ? true : false;

  if(seekable) { /* setup position tracking variables */
    /* total chunks into bitstream (if seekable) */
    frametot = server->gettotalframe();
    /* size of a chunk in bytes */
    framesize = server->getpcmperframe()<<2;
    /* how much frames make second */
    fps = samplerate/server->getpcmperframe();
  } else framesize = (1152*channels)<<1; /* hardcoded mp3 framesize in bytes */

  opened = true;
  quit = false;
  func("MpegChannel::set bitrate[%u] channels[%u]%sseekable",
       bitrate,channels,(seekable)?" ":" non ");
  
  unlock();

  return(res);
}

float MpegChannel::_get_audio() {
  float res;

  /*  uhm, i guess that's not needed
      if(!server->run(-1)) // Initialize MPEG Layer 3
      return(3.0);
  */

  if(!server->run(1)) {
    switch(server->geterrorcode()) {
    case SOUND_ERROR_FINISH:
      return(2.0);
      break;
    case SOUND_ERROR_FILEOPENFAIL:
      error("mpeglib: failed opening file (FILEOPENFAIL)");
      return(3.0);
      break;
    case SOUND_ERROR_FILEREADFAIL:
      error("mpeglib: failed reading from file (FILEREADFAIL)");
      return(3.0);
      break;
    case SOUND_ERROR_UNKNOWNPROXY:
      error("mpeglib: unknown proxy");
      return(3.0);
      break;
    case SOUND_ERROR_UNKNOWNHOST:
      error("mpeglib: unknown host");
      return(3.0);
      break;
    case SOUND_ERROR_SOCKET:
      error("mpeglib: socket error");
      return(3.0);
      break;
    case SOUND_ERROR_CONNECT:
      error("mpeglib: connect error");
      return(3.0);
      break;
    case SOUND_ERROR_FDOPEN:
      error("mpeglib: FDOPEN");
      return(3.0);
      break;
    case SOUND_ERROR_HTTPFAIL:
      error("mpeglib: http failure");
      return(3.0);
      break;
    case SOUND_ERROR_HTTPWRITEFAIL:
      error("mpeglib: http write failed");
      return(3.0);
      break;
    case SOUND_ERROR_TOOMANYRELOC:
      error("mpeglib: TOOMANYRELOC");
      return(3.0);
      break;
    case SOUND_ERROR_THREADFAIL:
      error("mpeglib: thread failure");
      return(3.0);
      break;
    default:
      error("mpeglib: unknown error :(");
      return(3.0);
      break;
    }
  }

  /* frame stores how much data we have in (in bytes)
     for mp3 it's fixed coz the framesize never changes
     (maybe with VBR does?) */
  frame = framesize; //server->getpcmperframe()<<2;

  if(seekable) {
    framepos = server->getcurrentframe();
    res = upd_time();
  } else res = 0.0;

  /* resampling deactivated (speed)
     if you are wondering: this is not enough for it.
     in_smp = resample((short int *)&server->rawdata[0],in_smp,buffo,speed);
  */

  /* writes into the pipe */
  write(frame,(IN_DATATYPE *)server->rawdata);

  return(res);
}

bool MpegChannel::pos(float pos) {

    if(pos==0.0)
      server->setframe(0);
    else
      server->setframe((int) (frametot*pos));
    /* no error codes out of here :( */

    time.f = position = pos;
  
  return(true);
}
